/*
 * Scrutation, mesure et impression du train d'impulsions du DCF77 type "Conrad"
 */

/*  Evolutions:
 *  1.0.5: Resserrement des durées des pulses ([70..130] mS et [170..230] mS)
 *  1.0.6: Formatage du train de bits avec séparation des champs
 *  1.0.7: Accueil des min, moyenne et max des durées des pulses '0' et '1' sur une trame correcte
 *  1.0.8: Add min, moyenne et max des pulses _/--\_ et -\__/- sur une trame incorrecte + struct use
 *  1.1.0: Accueil de l'auto adaptation des durées min et max des pulses '0 et '1' + la pause
 *  1.1.2: Add tests in 'setPulseAndPauseDurations()' method
 *  1.2.1: Accueil de la symbologie des 3 leds Verte, Orange et Rouge:
 *         - Led Verte:  o Clignotement à 1 Hz et allumage durant 200 mS si absence de pulse '0' et '1'
 *                         ou inactivité à l'état bas et haut durant plus de 3 Sec. (état "non connecté")
 *                       o Clignotement à 1 Hz et allumage durant 100 mS à la 1st  pulse '0' et '1' détectée (état "connecté")
 *                       o Extinction durant 1 Sec. sur détection de la phase de repos si dans l'état "connecté"
 *         - Led Orange: Permet de lire les 59 bits après la phase de repos (lecture visuelle à la fréquence de 1 Hz ;-)
 *                       o Extinction si pulse '0' ou phase de repos détectées
 *                       o Allumage si pulse '1' détectée                  
 *         - Led Rouge:  o Présentation des erreurs toutes les 6.4 Sec. (présentation durant au moins 1 minute jusqu'à
 *                         détection d'une nouvelle phase de repos)
 *                         - 1 allumage  de 400 mS si "Empty frame" ( __/-\__...)
 *                         - 2 allumages de 400 mS si "Frame too short" ( __/-\__/-\__...) (rapport cyclique 50%)
 *                         - 3 allumages de 400 mS si "Invalid frame"   ( __/-\__/-\__/-\__...) (rapport cyclique 50%)
 *                         - 4 allumages de 400 mS si "Frame too long"  ( __/-\__/-\__/-\__/-\__...) (rapport cyclique 50%)
 *                         - 5 allumages de 400 mS si "Secondes de synchro" > 1 Sec.
 *                         - 6 allumages de 400 mS si "Date et/ou temps" iestimés incorrectes > 1 Sec.
 *                         - 7 allumages de 400 mS (réserve)
 *                       o Allumage cyclique de 400 mS (rapport cyclique 50%) si "Internal error" sans limite de durée
 *                         pour s'assurer qu'aucune erreur interne ne se présente pendant tout le fonctionnement
 *                         (équivalent à 8 allumages ;-)
 *  1.2.2: Improving...
 *  1.3.1: Calcul, présentation et correction de la dérive des secondes comptabilisées en interne (+/- 1 Sec. tolérée)
 *  1.3.2: Détermination de la date et heure en mode "non connecté" au moyen de la seconde de synchro
 *  1.3.3: Affichage des statistiques de la gestion des timestamp
 *  1.3.4: Improving of presenttaion and formatage + get messages from 'Serial' for 'calculOfDateAndTime' method test...
 *  1.3.5: Add statistics:
 *         - Print if 'timestamp' > 'timestamp_max_max'
 *         - Availability in percent [0%..100%]: 100 * (Nbr total minutes - Nbr minutes calculated) / Nbr total minutes
 *  1.3.6: Add report des heures et minutes sur les 2 Leds LED_HOURS at LED_MINUTES
 *  1.3.7: Use calls by reference in 'verify()', 'printDateAndTime()' and 'printWithFormat' methods
 *         => Fix stack overflow (128 bytes max ;-)
 *  1.3.8: Add 'mktime' method for integrity test of next minute decoded (with test by 'Serial' use)
 *  1.3.9: Calcul and test of the "Current Epoch Time" and "Estimated Epoch Time" (+ 1 Mn @ "current")
 *  1.3.10: - Présentation de l'erreur "Current Epoch Time" diferent de "Estimated Epoch Time" (+/- 1 Mn @ "current")
 *          - Test avec le récepteur DCF77 de chez Reichelt (aucunne évolution logicielle @ Conrad ;-)
 *          - Correction pas de "Stats2" si le passage de la minute à 0 n'st pas "vu"
 *            => Changer "minute == MINUTE_FOR_STATS2" en "nouvelle heure"
 *
 *          TODO: - Si erreur interne (persistante), pas d'affichage des heures et minutes si autre erreur
 *                  mais affichage si trame décodée correcte ;-)
 *
 *  1.3.11: - Passage de 'TIMESTAMP_DELTA_MAX_MAX' à 200 mS
 *          - Fix affichage "Date and time calculated" si "#Invalid Epoch time [xxx] Sec. (Delta: [yyy] Sec.)"
 *  1.3.12: - Add duration of functionnement since the 1st correct frame
 */

#include <limits.h>     // For 'UINT_MAX', 'INT_MAX', etc. definitions

// Behaviour definitions
#define USE_PRINT_STATS          1
#define USE_PRINT_STATS2         1
#define USE_PRINT_STATS_TS       1
#define USE_GET_SERIAL_IN        0
#define USE_LEDS_FOR_TIME        1

// Prompt au lancement
#define PROMPT        "#UNO: DCF77 in progress (Version 1.3.12)..."

#if USE_GET_SERIAL_IN
size_t  g__count = 0;
char    g__incoming_buff[80 + 1];   // +1 for '\0' terminal
#endif

// Définitions pour la gestion des Leds
#define LED_YELLOW    12
#define LED_RED       11
#define LED_GREEN     13

#if USE_LEDS_FOR_TIME
#define LED_HOURS      2
#define LED_MINUTES    3
#endif

// Masques pour la recopie des états des leds
#define STATE_NO_LEDS          0
#define STATE_LED_RED         (1 << 0)
#define STATE_LED_YELLOW      (1 << 1)
#define STATE_LED_GREEN       (1 << 2)

int              g__state_leds = STATE_NO_LEDS;

// Définitions pour la gestion du DCF77
int              PIN_DCF77 = A3;
int              pulse_prev;
unsigned long    time_falling;
unsigned long    time_rising;
unsigned long    time_previous = 0L;

#define FRAME_LENGTH               59
#define TIMESTAMP_DELTA_MAX_MAX   200

enum {
  CURRENT = 0,             // Zone de travail (acquisition en cours)
  LAST_GOOD,               // Dernière trame et données associées sans erreur
  ESTIMATED,               // Date et heure estimées (calculée à partir de la dernière valide)
  NBR_ST_DATE_AND_TIME     // Nombre de structures pour leurs définitions
};

typedef struct {
  int     seconds;              // Seconde [0..58..(59)] du DCF77
  int     seconds_for_synchro;  // Seconde à l'expiration de 'st_gestion.duration_synchro'
  int     minutes;
  int     hours;
  int     day;
  int     day_of_week;
  int     month;
  int     year;
} ST_DATE_AND_TIME;

typedef struct {
  boolean             available;            // Infos disponibles
  unsigned long       timestamp;            // Timestamp des données acquises
  String              bits;                 // Résultat du train de bits détecté
  ST_DATE_AND_TIME    dt;                   // Date et temps calculés ou attendus
} ST_INFOS;

ST_INFOS g__infos[NBR_ST_DATE_AND_TIME];    // Last and expected date + time

boolean g__flg_sommer_hour  = true;        // Passage de l'heure d'été / hiver
long    g__epochTimeInit    = 0L;          // Epoch Time à l'initialisation (1st trame correcte)
long    g__epochTimeCurrent = 0L;          // Epoch Time courant
int     g__last_hour        = -1;          // Détection changement d'heure (pour "Stats2")

// Statistiques
enum {
  INFO_PULSE_0 = 0,
  INFO_PULSE_1,
  INFO_HIGHT,
  INFO_LOW,
  INFO_TIMESTAMP,
  INFO_NBR
};

enum {
  INFO_FRAME_OK = 0,
  INFO_FRAME_KO,
  INFO_TS
};

typedef struct {
  unsigned int samples;
  unsigned int min;
  unsigned int avg;
  unsigned int max;
  unsigned int out_of_range;
} ST_STATS;

// Statistiques never reseted
typedef struct {
  boolean       flg_report;
  unsigned long nbr_total_minutes;
  unsigned long nbr_minutes_calculated;
  int           derive_negative_max;
  int           derive_positive_max;
  unsigned int  timestamp_max_max;
} ST_STATS2;

ST_STATS     g__pulses[INFO_NBR];
ST_STATS2    g__stats2;
unsigned int g__pause_duration = 0;

// Définitions à l'init. des durées min et max des pulses '0 et '1' + la pause
// Values specified for the DCF77 from Reichelt ;-)
#define PULSE_0_MIN     40L
#define PULSE_0_MAX     130L
#define PULSE_1_MIN     140L
#define PULSE_1_MAX     230L
#define PAUSE_MIN       1500L

unsigned long g__pulse_0_min = PULSE_0_MIN;
unsigned long g__pulse_0_max = PULSE_0_MAX;
unsigned long g__pulse_1_min = PULSE_1_MIN;
unsigned long g__pulse_1_max = PULSE_1_MAX;
unsigned long g__pause_min   = PAUSE_MIN;

long          g__epoch_time_exp_delta_min = LONG_MAX;
long          g__epoch_time_exp_delta_max = LONG_MIN;

// Gestion des états, des erreurs et des timers pour la symbologie des Leds
typedef enum {
  NO_ERROR = 0,

  // Erreurs non persistantes
  EMPTY_FRAME,                         // Erreurs effaçables à chaque trame correcte détectée
  FRAME_TOO_SHORT,
  INVALID_FRAME,
  FRAME_TOO_LONG,
  SECOND_SYNCHRO_OUT_OF_RANGE,          // Dérive des secondes trop importante (> 1 Sec. en mode "connecté";-)
  INVALID_DATE_TIME_VS_ESTIMATED,       // Date et/ou temps invalides par rapport à la date et le temps estimée par 'my_mktime'

  // Erreurs persistantes
  INTERNAL_ERROR,                       // Marquage des erreurs internes non effaçables
  TS_100MS_OVERFLOW = INTERNAL_ERROR,   // TS > 100 mS en fond de tâche (64 mS max constatés dans la pratique)
} ERRORS;

typedef enum {
  UNKNOWN = 0,                  // 'time_pre' pas encore comptabilisé (état initial et unique)
  CONNECTED,                    // Expiration de 'duration_3sec' dans l'état 'CONNECTED_IN_PROGRESS'
  NOT_CONNECTED_IN_PROGRESS,    // Absence d'une pulse '0' ou '1' => armement de 'duration_3sec' si 'CONNECTED'
  NOT_CONNECTED                 // Expiration de 'duration_3sec' dans l'état 'NOT_CONNECTED_IN_PROGRESS'
} STATES;

typedef enum {
  NOT_DETECTED = 0,             // Etat de la ligne inconnu
  PULSE_0_DETECTED,             // Pulse '0' détectée
  PULSE_1_DETECTED,             // Pulse '1' détectée
  FRAME_END_DETECTED            // Finde trame détectée (60ième seconde)
} STATES_LINE;

/* Timerss géré en mS et initialisés à la valeur ad hoc
 *  o -1: Non armé
 *  o  0: Expiration
 *  o >0: Armé
 */
typedef struct {
  unsigned long time_previous;
  STATES        state;
  ERRORS        error;
  STATES_LINE   state_line;             // Etat de la ligne (détection pulse '0', '1' ou repos)

  int           nbr_for_error_pulses_init;
  int           nbr_for_error_pulses_work;

  int           duration_10mS;          // Comptabilise la durée pour la gestion des durée à cette résolution de 10 mS
  int           duration_connect;       // Comptabilise les durées pour les états "connecté" / "non connecté"
  int           duration_error_freq;    // Comptabilise le temps cyclique de présentation de l'erreur (fréquence)
  int           duration_error_pulse;   // Comptabilise les durées pour les présentations de l'erreurs (pulse)
  int           duration_synchro;       // Synchronisation avec le DCF77
  int           duration_synchro_pulse; // Présentation de la seconde de synchro (200 mS - 50% __/-\__/-\__... sur la led Verte)
  int           duration_frame_end;     // Présentation de la détection de la phase de repos
  int           duration_3Sec_pulse_0;  // Comptabilise les 3 Sec sur pulse à l'état bas  - TODO: Change name (line)
  int           duration_3Sec_pulse_1;  // Comptabilise les 3 Sec sur pulse à l'état haut - TODO: Change name (line)
  int           duration_3Sec;          // Comptabilise les 3 Sec

#if USE_LEDS_FOR_TIME
  int           duration_for_led_hours;     // Initialisé à 'hours' x 500 mS
  int           duration_for_led_minutes;   // Initialisé à 'minutes' x 500 mS
#endif
} ST_GESTION;

ST_GESTION      st_gestion;

// Fin des définitions constantes et variables globales

// Extract from '/usr/share/sdcc/lib/src/time.c'
#define LEAP_YEAR(year) ((year % 4) == 0)

typedef struct {
  int tm_sec;    /* Seconds. [0-60] (1 leap second) */
  int tm_min;    /* Minutes. [0-59] */
  int tm_hour;   /* Hours. [0-23] */
  int tm_mday;   /* Day.   [1-31] */
  int tm_mon;    /* Month. [0-11] */
  int tm_year;   /* Year - 1900.  */
} ST_TM;

/* Convert broken time to calendar time (seconds since 1970)
 *
 * Warning: Date limite en 2038 ;-)
*/
long my_mktime(ST_TM *timeptr)
{
static char monthDays[] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};

  int  year  = timeptr->tm_year + 1900;
  int  month = timeptr->tm_mon;
  int  i;
  long seconds;

  // Seconds from 1970 till 1 jan 00:00:00 this year
  seconds = (long)(year - 1970) * 60L * 60 * 24 * 365;  // seconds is a long

  // Add extra days for leap years
  for (i = 1970; i < year; i++) {
    if (LEAP_YEAR(i)) {
      seconds += 60L * 60* 24;        // seconds is a long
    }
  }

    // Add days for this year
  for (i = 0; i < month; i++) {
    if (i == 1 && LEAP_YEAR(year)) { 
      seconds += 60L * 60 * 24 * 29;  // seconds is a long
    }
    else {
      seconds += (long)monthDays[i] * 60L * 60 * 24;  // seconds is a long
    }
  }

  seconds += (long)(timeptr->tm_mday - 1) * 60L * 60 * 24;  // seconds is a long
  seconds += (long)timeptr->tm_hour * 60L * 60;
  seconds += (long)timeptr->tm_min * 60L;
  seconds += (long)timeptr->tm_sec;

  return seconds;
}
// Fin: Extract from '/usr/share/sdcc/lib/src/time.c'

long calculatedEpochTime(ST_DATE_AND_TIME *i__st_date_and_time) {

    ST_TM l__st_tm;
    memset(&l__st_tm, '\0', sizeof(ST_TM));

    l__st_tm.tm_mday = i__st_date_and_time->day;
    l__st_tm.tm_mon  = i__st_date_and_time->month - 1;
    l__st_tm.tm_year = i__st_date_and_time->year + (2000 - 1900);
    l__st_tm.tm_hour = i__st_date_and_time->hours;
    l__st_tm.tm_min  = i__st_date_and_time->minutes;

    // Application heure d'été / hiver
    return my_mktime(&l__st_tm) - (g__flg_sommer_hour ? 2L * 3600L : 3600L);
}

void clearStats() {
  unsigned int i = 0;
  for (i = 0; i < sizeof(g__pulses) / sizeof(g__pulses[0]); i++) {
    g__pulses[i].samples      = 0;
    g__pulses[i].min          = UINT_MAX;    // Pour provoquer une 1st initialisation après le test 'dt < min'
    g__pulses[i].avg          = 0;
    g__pulses[i].max          = 0;
    g__pulses[i].out_of_range = 0;
  }

  g__pause_duration = 0;
}

void clearStatsTimestamp() {
    g__pulses[INFO_TIMESTAMP].samples        = 0;
    g__pulses[INFO_TIMESTAMP].min            = UINT_MAX;    // Pour provoquer une 1st initialisation après le test 'dt < min'
    g__pulses[INFO_TIMESTAMP].avg            = 0;
    g__pulses[INFO_TIMESTAMP].max            = 0;
}

void setPulseAndPauseDurations() {
  char l__buffer[64];
  boolean l__flg = false;

  g__pulse_0_min = g__pulses[INFO_PULSE_0].min - (g__pulses[INFO_PULSE_0].min / 10);    // New value at -10%
  g__pulse_0_max = g__pulses[INFO_PULSE_0].max + (g__pulses[INFO_PULSE_0].max / 10);    // New value at +10%
  g__pulse_1_min = g__pulses[INFO_PULSE_1].min - (g__pulses[INFO_PULSE_1].min / 10);    // New value at -10%
  g__pulse_1_max = g__pulses[INFO_PULSE_1].max + (g__pulses[INFO_PULSE_1].max / 10);    // New value at +10%
  g__pause_min   = g__pause_duration - (g__pause_duration / 10);                        // New value at -10%          

  // Test des nouvelles valeurs des durée des pulses attendues
  // 1 - Test si 'g__pulse_0_min' < PULSE_0_MIN; auquel cas réinit. aux valeurs par défaut
  if (g__pulse_0_min < PULSE_0_MIN) {
    g__pulse_0_min = PULSE_0_MIN;
    l__flg = true;
  }

  // 2 - Test si 'g__pulse_0_max' proche de 'g__pulse_1_min'(< 10 mS); auquel cas réinit. aux valeurs par défaut
  if (g__pulse_0_max > (g__pulse_1_min - 10)) {
    g__pulse_0_max = PULSE_0_MAX;
    g__pulse_1_min = PULSE_1_MIN;
    l__flg = true;
  }

  // 3 - Test si 'g__pulse_1_max' > PULSE_1_MAX; auquel cas réinit. aux valeurs par défaut
  if (g__pulse_1_max > PULSE_1_MAX) {
    g__pulse_1_max = PULSE_1_MAX;
    l__flg = true;
  }

  // 4 - Test si g__pause_min < PAUSE_MIN; auquel cas réinit. à la valeur par défaut
  if (g__pause_min < PAUSE_MIN) {
    g__pause_min = PAUSE_MIN;
    l__flg = true;
  }
  // Fin: Test des nouvelles valeurs des durée des pulses attendues

  // Warning: à 9600 bauds, empiète sur la durée d'acquisition du 1st bit ;-(
  if (g__infos[CURRENT].dt.minutes == 0) {
    if (l__flg == true) {
      sprintf(l__buffer, "\n#New duration values: [0: %ld/%ld] [1: %ld/%ld] [P: %ld]", g__pulse_0_min, g__pulse_0_max, g__pulse_1_min, g__pulse_1_max, g__pause_min);
      Serial.print(l__buffer);
    }
  }
}

void setup() {
  char l__buffer[80];

  // Initialize digital pin LEDs as an output.
  pinMode(LED_RED, OUTPUT);
  pinMode(LED_YELLOW, OUTPUT);
  pinMode(LED_GREEN, OUTPUT);

#if USE_LEDS_FOR_TIME
  pinMode(LED_HOURS, OUTPUT);
  pinMode(LED_MINUTES, OUTPUT);
#endif

  // Test the LEDs on before processing
  digitalWrite(LED_RED, HIGH);
  digitalWrite(LED_YELLOW, HIGH);
  digitalWrite(LED_GREEN, HIGH);

#if USE_LEDS_FOR_TIME
  digitalWrite(LED_HOURS, HIGH);
  digitalWrite(LED_MINUTES, HIGH);
#endif

  delay(1000);
  digitalWrite(LED_RED, LOW);
  delay(500);
  digitalWrite(LED_YELLOW, LOW);
  delay(500);
  digitalWrite(LED_GREEN, LOW);

#if USE_LEDS_FOR_TIME
  delay(500);
  digitalWrite(LED_HOURS, LOW);
  delay(500);
  digitalWrite(LED_MINUTES, LOW);
#endif

  Serial.begin(19200);
  Serial.println(PROMPT);

#if USE_GET_SERIAL_IN
  Serial.println("#Warning: 'Serial.read()' for tests");
  sprintf(l__buffer, "#sizeof(int)  [%d] bytes\n", sizeof(int));
  Serial.print(l__buffer);
  sprintf(l__buffer, "#sizeof(long) [%d] bytes\n", sizeof(long));
  Serial.print(l__buffer);
#endif

  pinMode(PIN_DCF77, INPUT_PULLUP);

  pulse_prev = digitalRead(PIN_DCF77);
  time_falling = millis();
  time_rising = millis();

  clearStats();

  sprintf(l__buffer, "#Duration values (mS): [0: %ld/%ld] [1: %ld/%ld] [P: %ld] [TS Max-max: %d]",
    g__pulse_0_min, g__pulse_0_max, g__pulse_1_min, g__pulse_1_max, g__pause_min, TIMESTAMP_DELTA_MAX_MAX);
  Serial.println(l__buffer);

  // Reset the all fields of 'g__infos' structure
  size_t n = 0;
  for (n = 0; n < sizeof(g__infos) / sizeof(g__infos[CURRENT]); n++) {
    g__infos[n].timestamp = (unsigned long)-1;
    memset(&g__infos[n].dt, '\0', sizeof(ST_DATE_AND_TIME));
    g__infos[n].dt.seconds = -1;
    g__infos[n].bits = "";
    g__infos[n].available = false;
  }

  st_gestion.state = UNKNOWN;
  g__state_leds = STATE_NO_LEDS;

  memset(&g__stats2, '\0', sizeof(ST_STATS2));
  g__stats2.flg_report = false;
}

boolean parity(String str, char parity_bit) {
  unsigned int i;
  int sum = 0;

  for (i = 0; i < str.length(); i++) {
    if (str[i] == '1') sum += 1;
  }
  return ((sum & 1) + '0' == parity_bit);
}

boolean verify(String &str) {
  if (str[0] != '0')  return false;         // 1st Bit de la trame (toujours à 0)
  if (((str[17] - '0') ^ (str[18] - '0')) == 0) return false;           // Bits Z1 et Z2 exclusifs (codage du CET eor CEST))
  if (str[20] != '1') return false;         // Bit de début de codage des informations horaires (toujours à 1)
  if (parity(str.substring(21, 28), str[28]) == false) return false;    // Test de parité sur les minutes
  if (parity(str.substring(29, 35), str[35]) == false) return false;    // Test de parité sur les heures
  if (parity(str.substring(36, 58), str[58]) == false) return false;    // Test de parité sur la date (jour, mois et année)

  return true;
}

int decode_bcd(String str) {
  unsigned int i;
  int power_of_two;
  int result = 0;
  int power_of_ten = 1;

  for (i = 0; i < str.length(); i++) {
    power_of_two = (1 << (i & 0x3));
    result += (str[i] - '0') * power_of_two * power_of_ten;
    if (power_of_two == 8) power_of_ten *= 10;
  }

  return result;
}

boolean decode_time(String &str) {

  if (verify(str) == false) return false;
  g__infos[CURRENT].dt.minutes     = decode_bcd(str.substring(21, 28));    // Extraction of bits #21 to #27 not included -> Range [0..59] (field with parity)
  g__infos[CURRENT].dt.hours       = decode_bcd(str.substring(29, 35));    // Extraction of bits #29 to #35 not included -> Range [0..23] (field with parity)
  g__infos[CURRENT].dt.day         = decode_bcd(str.substring(36, 42));    // Extraction of bits #36 to #42 not included -> Range [1..31] (field without parity)
  g__infos[CURRENT].dt.day_of_week = decode_bcd(str.substring(42, 45));    // Extraction of bits #42 to #45 not included -> Range [1..7]  (1: Lundi, 2: Mardi, ..., 7: Dimanche)
  g__infos[CURRENT].dt.month       = decode_bcd(str.substring(45, 50));    // Extraction of bits #45 to #50 not included -> Range [1..12] (field without parity)
  g__infos[CURRENT].dt.year        = decode_bcd(str.substring(50, 58));    // Extraction of bits #50 to #58 not included -> Range [0..99] (field without parity)

  // Test des plages des données décodées (champs de la date jamais égaux à zéro ;-)
  if (g__infos[CURRENT].dt.minutes > 59)                                            return false;
  if (g__infos[CURRENT].dt.hours > 23)                                              return false;
  if (g__infos[CURRENT].dt.day < 1 || g__infos[CURRENT].dt.day > 31)                return false;
  if (g__infos[CURRENT].dt.day_of_week < 1 || g__infos[CURRENT].dt.day_of_week > 7) return false;
  if (g__infos[CURRENT].dt.month < 1 || g__infos[CURRENT].dt.month > 12)            return false;
  if (g__infos[CURRENT].dt.year < 19)                                               return false;    // Programme écrit en 2019 ;-)

  /* Update heure d'été / hiver si la trame décodée est entièrement correcte
   * Remarque: 'g__flg_sommer_hour' sera utilisée dans "Date and time calculated" estimée ;-)
   */
  g__flg_sommer_hour = (str[17] == '1') ? true : false;  // Test of bits #17 and #18 already processed

  // Détection du changement d'heure si la trame décodée est entièrement correcte
  if (g__last_hour != -1 && g__last_hour != g__infos[CURRENT].dt.hours) {

    // Print 'stats2' toutes les heures
    g__stats2.flg_report = true;
  }
  g__last_hour = g__infos[CURRENT].dt.hours;

  return true;
}

String getDayOfWeek(int num_day) {
  switch(num_day) {
  case 1:  return "Lundi";
  case 2:  return "Mardi";
  case 3:  return "Mercredi";
  case 4:  return "Jeudi";
  case 5:  return "Vendredi";
  case 6:  return "Samedi";
  case 7:  return "Dimanche";
  default: return "Unknown";
  }
}

boolean printDateAndTime(String &str_frame) {
  char l__buffer[16];

  if (decode_time(str_frame) == true) {
    Serial.print("[");
    Serial.print(getDayOfWeek(g__infos[CURRENT].dt.day_of_week));
    sprintf(l__buffer, " %02d/%02d/%d", g__infos[CURRENT].dt.day, g__infos[CURRENT].dt.month, 2000+ g__infos[CURRENT].dt.year);
    Serial.print(l__buffer);
    sprintf(l__buffer, " %02d:%02d", g__infos[CURRENT].dt.hours, g__infos[CURRENT].dt.minutes);
    Serial.print(l__buffer);
    Serial.print(" (Heure d'");
    if (g__flg_sommer_hour) {
      Serial.print("été)");
    }
    else {
      Serial.print("hiver)");
    }
    Serial.print("]");

    if (st_gestion.error < INTERNAL_ERROR) {   // "Internal error" persistante ;-)
      st_gestion.error = NO_ERROR;
      st_gestion.nbr_for_error_pulses_init = 0;
    }
  }
  else if (str_frame.length() == 0) {
   if (st_gestion.error < INTERNAL_ERROR) {   // "Internal error" persistante ;-)
     st_gestion.error = EMPTY_FRAME;
     st_gestion.nbr_for_error_pulses_init = 1;    // EMPTY_FRAME
   }

    Serial.print("Error: Empty frame");
    return false;
  }
  else if (str_frame.length() < FRAME_LENGTH) {
   if (st_gestion.error < INTERNAL_ERROR) {   // "Internal error" persistante ;-)
     st_gestion.error = FRAME_TOO_SHORT;
     st_gestion.nbr_for_error_pulses_init = 2;    // FRAME_TOO_SHORT
   }

    Serial.print("Error: Frame too short");
    return false;
  }
  else {
   if (st_gestion.error < INTERNAL_ERROR) {   // "Internal error" persistante ;-)
     st_gestion.error = INVALID_FRAME;
     st_gestion.nbr_for_error_pulses_init = 3;    // INVALID_FRAME
   }

    Serial.print("Error: Invalid frame");
    return false;
  }

  return true;
}

void printWithFormat(String &str_frame) {
  char l__buffer[FRAME_LENGTH+1+8+1];   // +1 for frame too long, +8 for separators '-' and +1 for '\0' reservations
  unsigned int i, j;

  for (i = j = 0; i < str_frame.length() && j < sizeof(l__buffer); i++) {
    switch (i) {
    case 1:
    case 15:
    case 21:
    case 29:
    case 36:
    case 42:
    case 45:
    case 50:
      l__buffer[j++] = '-';
      break;

    default:
      break;
    }
    l__buffer[j++] = str_frame[i];
  }

  l__buffer[j] = '\0';
  Serial.print(l__buffer);
}

void printDateAndTime(void) {
  char l__buffer[16];

  Serial.print(getDayOfWeek(g__infos[ESTIMATED].dt.day_of_week));
  sprintf(l__buffer, " %02d/%02d/%d", g__infos[ESTIMATED].dt.day, g__infos[ESTIMATED].dt.month, 2000+ g__infos[ESTIMATED].dt.year);
  Serial.print(l__buffer);
  sprintf(l__buffer, " %02d:%02d", g__infos[ESTIMATED].dt.hours, g__infos[ESTIMATED].dt.minutes);
  Serial.print(l__buffer);
}

#if USE_PRINT_STATS
void printStats(int l__num_stats) {
  char l__buffer[48];

  if (l__num_stats == INFO_FRAME_OK) {
    sprintf(l__buffer, " [0: #%d:%d/%d/%d]",
      g__pulses[INFO_PULSE_0].samples, g__pulses[INFO_PULSE_0].min, g__pulses[INFO_PULSE_0].avg, g__pulses[INFO_PULSE_0].max);
    Serial.print(l__buffer);
    sprintf(l__buffer, " [1: #%d:%d/%d/%d]",
      g__pulses[INFO_PULSE_1].samples, g__pulses[INFO_PULSE_1].min, g__pulses[INFO_PULSE_1].avg, g__pulses[INFO_PULSE_1].max);
    Serial.print(l__buffer);
    sprintf(l__buffer, " [P: %d]", g__pause_duration);
    Serial.print(l__buffer);
    sprintf(l__buffer, " [OOR: #%d]", g__pulses[INFO_HIGHT].out_of_range);
    Serial.print(l__buffer);
  }
  else if (l__num_stats == INFO_FRAME_KO) {
    sprintf(l__buffer, " [H: #%d:%d/%d]",
      g__pulses[INFO_HIGHT].samples, g__pulses[INFO_HIGHT].min, g__pulses[INFO_HIGHT].max);
    Serial.print(l__buffer);
    sprintf(l__buffer, " [L: #%d:%d/%d]",
      g__pulses[INFO_LOW].samples, g__pulses[INFO_LOW].min, g__pulses[INFO_LOW].max);
    Serial.print(l__buffer);
    sprintf(l__buffer, " [OOR: #%d]", g__pulses[INFO_HIGHT].out_of_range);
    Serial.print(l__buffer);
  }
#if USE_PRINT_STATS_TS
  else if (l__num_stats == INFO_TS) {
    sprintf(l__buffer, " [TS Max: %d mS]", g__pulses[INFO_TIMESTAMP].max);
    Serial.print(l__buffer);
  }
#endif
}
#endif

#if USE_PRINT_STATS2
void printStats2() {
  char l__buffer[64];

  // Calcul de la durée de fonctionnement
  long l__duration    = (g__epochTimeCurrent - g__epochTimeInit);
  int  l__nbr_days    = (int)(l__duration / 86400L);
  int  l__nbr_hours   = (int)((l__duration % 86400L) / 3600L);
  int  l__nbr_minutes = (int)(((l__duration % 86400L) % 3600L) / 60L);
  sprintf(l__buffer, "#Stats2: [Duration: %dJ %dH %d']", l__nbr_days, l__nbr_hours, l__nbr_minutes);
  Serial.print(l__buffer);

  sprintf(l__buffer, " [TS Max-max: %d mS]", g__stats2.timestamp_max_max);
  Serial.print(l__buffer);
  sprintf(l__buffer, " [Nbr minutes calculated/total: %ld/%ld", g__stats2.nbr_minutes_calculated, g__stats2.nbr_total_minutes);
  Serial.print(l__buffer);

  // Calcul de la disponiblité en pourcentage [0%..100%]: 100 * (Nbr total minutes - Nbr minutes calculated) / Nbr total minutes
  if (g__stats2.nbr_total_minutes > 0) {
    unsigned long l__availability = 100 * (g__stats2.nbr_total_minutes - g__stats2.nbr_minutes_calculated);
    l__availability /= g__stats2.nbr_total_minutes;

    sprintf(l__buffer, " (Availability: %d%%)]", (int)l__availability);
    Serial.print(l__buffer);
  }
  else {
    Serial.print("]");
  }
  // Fin: Calcul de la disponiblité en pourcentage

  sprintf(l__buffer, " [Derive Max: %d/+%d]", g__stats2.derive_negative_max, g__stats2.derive_positive_max);
  Serial.print(l__buffer);

  sprintf(l__buffer, " [Delta Epoch Time Min/Max: %ld/+%ld]", g__epoch_time_exp_delta_min, g__epoch_time_exp_delta_max);
  Serial.print(l__buffer);

  Serial.print("\n");
}
#endif

/*  Calcul de la prochaine date et heure à partir de la dernière valide (g__infos[LAST_GOOD] ou g__infos[ESTIMATED])
 *  Remarques :
 *   - Test de l'année bissextile limité à la division par 4
 *   - Passage de l'heure été / hiver non calculé
 */
void calculOfDateAndTime() {

  if (g__infos[LAST_GOOD].available == true && g__infos[ESTIMATED].available == false) {
    memcpy(&g__infos[ESTIMATED].dt, &g__infos[LAST_GOOD].dt, sizeof(ST_DATE_AND_TIME));
  }

  /* Disponibilité de 'g__infos[ESTIMATED]' pour le prochain calcul
   * Remarque: Sera marqué indisponible lorsque 'g__infos[LAST_GOOD]' sera de nouveau maj
   *           sur la réception d'une trame correcte
   */
  g__infos[ESTIMATED].available = true;

  g__infos[ESTIMATED].dt.seconds_for_synchro = -1;    // Maj mais non utilisé
  g__infos[ESTIMATED].dt.seconds = 0;                 // Maj mais non utilisé

  if (g__infos[ESTIMATED].dt.minutes < 59) {
    g__infos[ESTIMATED].dt.minutes += 1;
    goto rtn;   // Because test of new date and time
  }
  else if (g__infos[ESTIMATED].dt.hours < 23) {
    // Nouvelle heure
    g__infos[ESTIMATED].dt.minutes = 0;
    g__infos[ESTIMATED].dt.hours  += 1;

    // Print 'stats2' toutes les heures
    g__stats2.flg_report = true;

    goto rtn;   // Because test of new date and time
  }
  else {
    // Nouvelle heure
    // Print 'stats2' toutes les heures
    g__stats2.flg_report = true;

    // Nouvelle journée
    int l__last_day = -1;    // Dernier jour du mois
 
    g__infos[ESTIMATED].dt.minutes     = 0;
    g__infos[ESTIMATED].dt.hours       = 0;
    g__infos[ESTIMATED].dt.day_of_week = (g__infos[ESTIMATED].dt.day_of_week < 7) ? (g__infos[ESTIMATED].dt.day_of_week + 1) : 0;

    switch (g__infos[ESTIMATED].dt.month) {
    case 4:   // Mois de 30 jours
    case 6:
    case 9:
    case 11:
      l__last_day = 30;
      break;

    case 1:   // Mois de 31 jours
    case 3:
    case 5:
    case 7:
    case 8:
    case 10:
    case 12:
      l__last_day = 31;
      break;

    case 2:   // Mois de février
      /* Année bissextile:
       *    - Si l'année est divisible par 4 (testé) et non divisible par 100 (non testé - 2100 ... très loin ;-)
       *    - Si l'année est divisible par 400 (non testé - 2400 ... encore plus loin ;-)
       */
      l__last_day = (((2000 + g__infos[ESTIMATED].dt.year) % 4) == 0) ? 29 : 28;
      break;
    }

    if (g__infos[ESTIMATED].dt.day < l__last_day) {
      g__infos[ESTIMATED].dt.day += 1;
      goto rtn;   // Because test of new date and time
    }
    else {
      // Nouveau mois
      g__infos[ESTIMATED].dt.day = 1;

      if (g__infos[ESTIMATED].dt.month < 12) {
        g__infos[ESTIMATED].dt.month++;
        goto rtn;   // Because test of new date and time
      }
      else {
        // Nouvelle année
        g__infos[ESTIMATED].dt.month = 1;
        g__infos[ESTIMATED].dt.year++;
      }
    }
  }

rtn:
  g__last_hour = g__infos[ESTIMATED].dt.hours;    // Sauvegarde de l'heure calculée

#if USE_LEDS_FOR_TIME
  if (st_gestion.duration_for_led_hours <= 0 && st_gestion.duration_for_led_minutes <= 0) {   // // Pas de réarmement si en cours
    if (g__infos[ESTIMATED].dt.hours > 0) {
      // -1 pour le modulo 50 (2 plages [49, 48, ..., 24, 25]  et [24, 23, ..., 1, 0])
      st_gestion.duration_for_led_hours = 100 * g__infos[ESTIMATED].dt.hours - 1;       //  ~1 Sec. x hours
    }
    if (g__infos[ESTIMATED].dt.minutes > 0) {
      // -1 pour le modulo 50 (2 plages [49, 48, ..., 24, 25]  et [24, 23, ..., 1, 0])
      st_gestion.duration_for_led_minutes = 100 * g__infos[ESTIMATED].dt.minutes - 1;   //  1 Sec. x minutes
    }

    // Calcul of Epoch Time
    long l__seconds = calculatedEpochTime(&g__infos[ESTIMATED].dt);
    char l__buffer[80];

    if (g__epochTimeCurrent > 0L) {
      long l__delta = (g__epochTimeCurrent + 60L - l__seconds);
      g__epoch_time_exp_delta_min = (l__delta < g__epoch_time_exp_delta_min) ? l__delta : g__epoch_time_exp_delta_min;
      g__epoch_time_exp_delta_max = (l__delta > g__epoch_time_exp_delta_max) ? l__delta : g__epoch_time_exp_delta_max;

      //sprintf(l__buffer, "#Epoch time [%ld] Sec. (Delta: [%ld] Sec.)\n", l__seconds, l__delta);
      //Serial.print(l__buffer);

      if (l__delta != 0L) {
        if (st_gestion.error < INTERNAL_ERROR) {   // "Internal error" persistante ;-)
          st_gestion.error = INVALID_DATE_TIME_VS_ESTIMATED;
          st_gestion.nbr_for_error_pulses_init = 6;    // INVALID_DATE_TIME_VS_ESTIMATED

          sprintf(l__buffer, "#Invalid Epoch time [%ld] Sec. (Delta: [%ld] Sec.)\n", l__seconds, l__delta);
          Serial.print(l__buffer);
        }
      }
    }
    else {
      sprintf(l__buffer, "#Epoch time [%ld] Sec.\n", l__seconds);
      Serial.print(l__buffer);

      g__epochTimeInit = l__seconds;          // 1st Epoch Time calculated
    }

    g__epochTimeCurrent = l__seconds;
  }
#endif
}

void statesGestion(unsigned long i__time) {
  if (st_gestion.state == UNKNOWN) {
    st_gestion.state = NOT_CONNECTED;
    st_gestion.error = NO_ERROR;
    st_gestion.nbr_for_error_pulses_init = 0;
    st_gestion.nbr_for_error_pulses_work = 0;

    st_gestion.duration_10mS          = 10;
    st_gestion.duration_connect       = 100;          // 1 Sec
    st_gestion.duration_synchro       = -1;
    st_gestion.duration_error_freq    = -1;
    st_gestion.duration_error_pulse   = -1;
    st_gestion.duration_3Sec_pulse_0  = -1;
    st_gestion.duration_3Sec_pulse_1  = -1;
    st_gestion.duration_3Sec          = -1;
    st_gestion.duration_frame_end     = -1;
    st_gestion.duration_synchro_pulse = -1;

#if USE_LEDS_FOR_TIME
    st_gestion.duration_for_led_hours   = -1;
    st_gestion.duration_for_led_minutes = -1;
#endif
  }
  else {
    int l__delta_time = (int)(i__time - st_gestion.time_previous);    // Comptabilisation des mS écoulées depuis la dernière fois

    // Statistiques de la gestion des timestamp (seul le max est déteminé)
    if ((unsigned int)l__delta_time > g__pulses[INFO_TIMESTAMP].max) g__pulses[INFO_TIMESTAMP].max = (unsigned int)l__delta_time;
    // Fin: Statistiques de la gestion des timestamp

    /* Test du débordement de 'i__time' défini sur un 'unsigned long'
     *    32 bits non signés =>  4294967295 mS = 4294967 Sec. = ~49 jours et 17 heures
     */
    if (l__delta_time <= 0) {
      return;
    }

    /* 'l__delta_time' peut être supérieur à 1 mS dans le cas d'une surcharge de travail ;-)
     *  Remarques:
     *     - "Internal error" si néanmoins > TIMESTAMP_DELTA_MAX_MAX mS (à adapter suivant les circonstances)
     *     - La seule conséquence est une perte de n * 10 mS suite au test 'st_gestion.duration_10mS <= 0' ;-)
     */
    if (l__delta_time > TIMESTAMP_DELTA_MAX_MAX) {
      char l__buffer[48];

      sprintf(l__buffer, "#Timestamp overflow %d > %d\n", l__delta_time, TIMESTAMP_DELTA_MAX_MAX);
      Serial.print(l__buffer);

      st_gestion.error = TS_100MS_OVERFLOW;
      st_gestion.nbr_for_error_pulses_init = 8;   // TS_100MS_OVERFLOW
    }

    st_gestion.duration_10mS -= l__delta_time;

    if (st_gestion.duration_10mS <= 0) {
      // Décrémentation des durées de tous les timer armés (résolution de 10 mS)
      if (st_gestion.duration_connect       > 0) st_gestion.duration_connect--;
      if (st_gestion.duration_synchro       > 0) st_gestion.duration_synchro--;
      if (st_gestion.duration_error_pulse   > 0) st_gestion.duration_error_pulse--;
      if (st_gestion.duration_3Sec_pulse_0  > 0) st_gestion.duration_3Sec_pulse_0--;
      if (st_gestion.duration_3Sec_pulse_1  > 0) st_gestion.duration_3Sec_pulse_1--;
      if (st_gestion.duration_frame_end     > 0) st_gestion.duration_frame_end--;
      if (st_gestion.duration_synchro_pulse > 0) st_gestion.duration_synchro_pulse--;
      if (st_gestion.duration_3Sec          > 0) st_gestion.duration_3Sec--;
      if (st_gestion.duration_error_freq    > 0) st_gestion.duration_error_freq--;

      st_gestion.duration_10mS = 10;    // Réarmement

#if USE_LEDS_FOR_TIME
      if (st_gestion.duration_for_led_hours   > 0) st_gestion.duration_for_led_hours--;
      if (st_gestion.duration_for_led_minutes > 0) st_gestion.duration_for_led_minutes--;
#endif
    }

    if (st_gestion.duration_frame_end <= 0) {
      st_gestion.duration_frame_end = -1;
    }

    if (st_gestion.duration_synchro_pulse <= 0) {
      st_gestion.duration_synchro_pulse = -1;
    }

    if (st_gestion.duration_connect <= 0) {
      st_gestion.duration_connect = 100;     // 1 Sec.
    }

    if (g__infos[LAST_GOOD].available == true && st_gestion.duration_synchro <= 0) {
      char l__buffer[48];

      /* Warning: Cette 'seconds_for_synchro' ne sera véritablement la seconde de synchro
       *          que si la trame courante est détectée comme correcte auquel cas un offset 
       *          de +/- xxx secondes = ('seconds_for_synchro' - 30 Sec.) est appliqué
       */
      if (st_gestion.state == CONNECTED && (st_gestion.error == NO_ERROR || st_gestion.error == INVALID_DATE_TIME_VS_ESTIMATED || st_gestion.error >= INTERNAL_ERROR)) {
        int l__derive = INT_MAX;
        boolean l__flg = false;

        g__infos[CURRENT].dt.seconds_for_synchro = g__infos[CURRENT].dt.seconds;
        l__derive = (g__infos[CURRENT].dt.seconds_for_synchro - 30);

        // Réarmement de la seconde de synchro avec l'application de la dérive
        st_gestion.duration_synchro = 6000 - 100 * l__derive;       // 1 Min - dérive ramenée en multiple de 10 mS

        if (l__derive < -1 || l__derive > 1) {
          sprintf(l__buffer, "#Seconds for synchro #%d (Derive: %d Sec.) ", g__infos[CURRENT].dt.seconds_for_synchro, l__derive);
          Serial.print(l__buffer);
          l__flg = true;

          if (st_gestion.error == NO_ERROR) {   // Erreurs internes persistantes
            st_gestion.error = SECOND_SYNCHRO_OUT_OF_RANGE;
            st_gestion.nbr_for_error_pulses_init = 5;   // SECOND_SYNCHRO_OUT_OF_RANGE
          }

          // Mémorisataion de la dérive maximale (négative et positive)
          if (l__derive < g__stats2.derive_negative_max) {
             g__stats2.derive_negative_max = l__derive;
          }
          else if (l__derive > g__stats2.derive_positive_max) {
             g__stats2.derive_positive_max = l__derive;
          }
        }

#if USE_PRINT_STATS
#if USE_PRINT_STATS_TS
        // Statistiques de la gestion des timestamp (affichage que si dernier max est > à celui mémorisé)
        if (g__pulses[INFO_TIMESTAMP].max > g__stats2.timestamp_max_max) {
          g__stats2.timestamp_max_max = g__pulses[INFO_TIMESTAMP].max;
          Serial.print("#Stats:");
          printStats(INFO_TS);
          l__flg = true;
        }
#endif
#endif
        clearStatsTimestamp();
        if (l__flg == true) Serial.print("\n");
      }
      else {
        /* Réarmement de la seconde de synchro sans application de la dérive (non calculable)
         * dans le l'état "non connecté" ou erreur
         */
        st_gestion.duration_synchro = 6000;       // 1 Min (dérive tant qu'aucune trame valide détectée ;-)

        // Calcul de la date complète et de heure à partir de la dernière disponible dans 'g__infos[ESTIMATED]'
        calculOfDateAndTime();

        // Nouvelle minute présentée
        g__stats2.nbr_total_minutes++;
        g__stats2.nbr_minutes_calculated++;

        Serial.print("[");
        printDateAndTime();
        Serial.print("] (Date and time calculated) ");

#if USE_PRINT_STATS
#if USE_PRINT_STATS_TS
        // Statistiques de la gestion des timestamp (affichage que si dernier max est > à celui mémorisé)
        if (g__pulses[INFO_TIMESTAMP].max > g__stats2.timestamp_max_max) {
          g__stats2.timestamp_max_max = g__pulses[INFO_TIMESTAMP].max;
          Serial.print("#Stats:");
          printStats(INFO_TS);
        }
#endif
#endif
        clearStatsTimestamp();
        Serial.print("\n");
      }

      // Présentation de cette seconde de synchronisation
      st_gestion.duration_synchro_pulse = 100;  // 1 Sec.
    }

    // Présentation de l'état 'connecté' ou 'non connecté'
    if (st_gestion.duration_connect > 90) {          // Rapport cyclique 0..90..100
      if (st_gestion.state == NOT_CONNECTED) {
        g__state_leds |= STATE_LED_GREEN;            // Allumages __/--\____....
      }
      else if (st_gestion.state == CONNECTED) {
        g__state_leds &= ~STATE_LED_GREEN;           // Allumages --\__/----....
      }
    }
    else {
      if (st_gestion.state == NOT_CONNECTED) {
        g__state_leds &= ~STATE_LED_GREEN;           // Allumages --\__/----....
      }
      else if (st_gestion.state == CONNECTED) {
        g__state_leds |= STATE_LED_GREEN;            // Allumages __/--\____....
      }
    }
    // Fin: Présentation de l'état 'connecté' ou 'non connecté'
  }

  // Gestion des transitions état 'connecté' -> 'non connecté' (hystérésis)
  if (st_gestion.duration_3Sec == 0) {
    if (st_gestion.state == NOT_CONNECTED_IN_PROGRESS) {
      st_gestion.state = NOT_CONNECTED;
   }
   st_gestion.duration_3Sec = -1;
  }

  if (st_gestion.duration_3Sec_pulse_0 == 0) {
    if (st_gestion.state == CONNECTED) {
      st_gestion.state = NOT_CONNECTED;
      g__state_leds &= ~STATE_LED_YELLOW;
    }
    st_gestion.duration_3Sec_pulse_0 = -1;
  }
  if (st_gestion.duration_3Sec_pulse_1 == 0) {
    if (st_gestion.state == CONNECTED) {
      st_gestion.state = NOT_CONNECTED;
      g__state_leds &= ~STATE_LED_YELLOW;
    }
    st_gestion.duration_3Sec_pulse_1 = -1;
  }
  // Fin: Gestion des transitions état 'connecté' <-> 'non connecté'

  // Recopie sur les 2 leds Orange et Rouge des états des pulses '0', '1' et phase de repos détectés
  switch (st_gestion.state_line) {
  case NOT_DETECTED:
  case PULSE_0_DETECTED:
  case FRAME_END_DETECTED:
    g__state_leds &= ~STATE_LED_YELLOW;
    break;
  case PULSE_1_DETECTED:
    g__state_leds |= STATE_LED_YELLOW;
    break;
  default:
    break;
  }

  if (st_gestion.duration_frame_end > 0) {
    // Présenttaion de la phase de repos
    g__state_leds &= ~STATE_LED_GREEN;
  }
  else if (st_gestion.duration_synchro_pulse > 0) {
    /* Présenttaion de la seconde de synchro si pas confondue avec la phase de repos
     * Remarque: Ne doit pas arriver dans l'état "connecté" car cela signifierait une dérive
     *           de +/- 30 Sec. ;-) 
     */
    if (((st_gestion.duration_synchro_pulse / 10) % 2) == 0) {
      g__state_leds &= ~STATE_LED_GREEN;
    }
    else {
      g__state_leds |= STATE_LED_GREEN;
    }
  }
  // Fin: Recopie sur les 2 leds Orange et Rouge des états des pulses '0', '1' et phase de repos détectés

  /* Présentation des pulses d'erreurs pour une lecture visuelle
   *  - Présentation toutes les 5 Sec. du train d'allumage 200 mS - 50% correspondant à:
   *    o 1 allumage  pour 'Frame too short'
   *    o 2 allumages pour 'Invalid frame'
   *    o 3 allumages pour 'Frame too long'
   *    o 4 allumages pour 'Internal error'
   */
  if (st_gestion.error != NO_ERROR) {
    if ((st_gestion.duration_error_freq == -1 && st_gestion.duration_error_pulse == -1)   // Initialisation du processus
      || st_gestion.duration_error_freq == 0)   // Réinitialisation du processus toutes les 5 Sec.
    {
      st_gestion.duration_error_freq = 640;    // 6.4 Sec. (multiple de 800 mS pour permettre un clignotement correct infini)
      st_gestion.duration_error_pulse = 80;    // dans le cas où 'nbr_for_error_pulses_init' >= 8 =  (800 mS / 400 mS)
      st_gestion.nbr_for_error_pulses_work = st_gestion.nbr_for_error_pulses_init;
    }

    if (st_gestion.duration_error_pulse == 0) {
      // Réarmement après __/--\__... correspondant à 1 allumage
      st_gestion.duration_error_pulse = 80;     // 80 mS
      st_gestion.nbr_for_error_pulses_work -= 1;      // -1 sur le nombre d'allumages initialisé au double souhaité
    }

    if (st_gestion.nbr_for_error_pulses_work > 0) {
      // Recopie de l'état du timer dans la led Rouge
      if (st_gestion.duration_error_pulse > 40) {    // 400 mS: Rapport cyclique 50%
        g__state_leds |= STATE_LED_RED;
      }
      else {
        g__state_leds &= ~STATE_LED_RED;
      }
    }
    else {
      // Fin de tous les allumages
      g__state_leds &= ~STATE_LED_RED;
    }
  }
  else if (st_gestion.nbr_for_error_pulses_init > 0) {
    // Fin de la présentation de la précédente erreur
    st_gestion.duration_error_freq = -1;
    st_gestion.duration_error_pulse = -1;
    st_gestion.nbr_for_error_pulses_init = 0;      
    st_gestion.nbr_for_error_pulses_work = 0;
  }
  else {
    g__state_leds &= ~STATE_LED_RED;
  }
  // Fin: Présentation des pulses d'erreurs

#if USE_LEDS_FOR_TIME
  if (st_gestion.duration_for_led_hours > 0) {
    // Rapport cynclique de 50% (Eteinte si [0..49] et Allumé si [50..99])
    digitalWrite(LED_HOURS, ((st_gestion.duration_for_led_hours % 100) > 50) ? HIGH : LOW);
  }
  else {
    digitalWrite(LED_HOURS, LOW);
  }
  if (st_gestion.duration_for_led_minutes > 0) {
    // Rapport cynclique de 50% (Eteinte si [0..49] et Allumé si [50..99])
    digitalWrite(LED_MINUTES, ((st_gestion.duration_for_led_minutes % 100) > 50) ? HIGH : LOW);
  }
  else {
    digitalWrite(LED_MINUTES, LOW);
  }
#endif

  st_gestion.time_previous = i__time;
}

void loop() {
  boolean       l__flg_frame = false;
  boolean       l__flg_transistion = false;
  unsigned long l__delta = 0L;
  unsigned long time = millis();

  int pulse = digitalRead(PIN_DCF77);

  // Détection de l'inactivité à l'état bas et haut
  if (pulse == 0) {
    if (st_gestion.duration_3Sec_pulse_0 == -1) {
      st_gestion.duration_3Sec_pulse_0 = 300;   // 3 Sec.
    }
    st_gestion.duration_3Sec_pulse_1 = -1;
  }
  else {
    if (st_gestion.duration_3Sec_pulse_1 == -1) {
      st_gestion.duration_3Sec_pulse_1 = 300;   // 3 Sec.
    }
    st_gestion.duration_3Sec_pulse_0 = -1;
  }
  // Fin: Détection de l'inactivité à l'état bas et haut

  if (pulse == 1 && pulse_prev == 0) {
    // Détection du front montant si DCF77 connecté directement à l'entrée PIN_DCF77
    time_rising = time;
    l__delta = (time_rising - time_falling);  // Durée entre les fronts --\__ et __/--

    if ((unsigned int)l__delta < g__pulses[INFO_LOW].min) g__pulses[INFO_LOW].min = (unsigned int)l__delta;
    if ((unsigned int)l__delta > g__pulses[INFO_LOW].max) g__pulses[INFO_LOW].max = (unsigned int)l__delta;
    g__pulses[INFO_LOW].samples++;

    l__flg_transistion = true;
  }
  else if (pulse == 0 && pulse_prev == 1) {
    // Détection du front descendant si DCF77 connecté directement à l'entrée PIN_DCF77
    time_falling = time;
    l__delta = (time_falling - time_rising);  // Durée entre les fronts __/-- et --\__

    if ((unsigned int)l__delta < g__pulses[INFO_HIGHT].min) g__pulses[INFO_HIGHT].min = (unsigned int)l__delta;
    if ((unsigned int)l__delta > g__pulses[INFO_HIGHT].max) g__pulses[INFO_HIGHT].max = (unsigned int)l__delta;
    g__pulses[INFO_HIGHT].samples++;

    l__flg_transistion = true;
  }

  pulse_prev = pulse;

  if (l__flg_transistion == false) {
    goto more;
  }

  // Traitement sur les durées uniquement si transition détectée
  if (l__delta > g__pulse_0_min) {
    if (l__delta < g__pulse_0_max) {
      g__infos[CURRENT].dt.seconds++;
      g__infos[CURRENT].bits += '0';    // Pulse '0' détectée

      if ((unsigned int)l__delta < g__pulses[INFO_PULSE_0].min) g__pulses[INFO_PULSE_0].min = (unsigned int)l__delta;
      if ((unsigned int)l__delta > g__pulses[INFO_PULSE_0].max) g__pulses[INFO_PULSE_0].max = (unsigned int)l__delta;
      g__pulses[INFO_PULSE_0].avg = (g__pulses[INFO_PULSE_0].samples * g__pulses[INFO_PULSE_0].avg + (unsigned int)l__delta) / (g__pulses[INFO_PULSE_0].samples + 1);
      g__pulses[INFO_PULSE_0].samples++;

      st_gestion.state = CONNECTED;
      st_gestion.state_line = PULSE_0_DETECTED;
   }
   else if (l__delta > g__pulse_1_min && l__delta < g__pulse_1_max) {
      g__infos[CURRENT].dt.seconds++;
      g__infos[CURRENT].bits += '1';    // Pulse '1' détectée

      if ((unsigned int)l__delta < g__pulses[INFO_PULSE_1].min) g__pulses[INFO_PULSE_1].min = (unsigned int)l__delta;
      if ((unsigned int)l__delta > g__pulses[INFO_PULSE_1].max) g__pulses[INFO_PULSE_1].max = (unsigned int)l__delta;
      g__pulses[INFO_PULSE_1].avg = (g__pulses[INFO_PULSE_1].samples * g__pulses[INFO_PULSE_1].avg + (unsigned int)l__delta) / (g__pulses[INFO_PULSE_1].samples + 1);
      g__pulses[INFO_PULSE_1].samples++;

      st_gestion.state = CONNECTED;
      st_gestion.state_line = PULSE_1_DETECTED;
    }
    else if (pulse == 0) {
      // Pulse _/--\__ détectée de durée > 'g__pulse_0_min' mais pas dans la plage [g__pulse_0_min..g__pulse_0_max] et [g__pulse_1_min..g__pulse_1_max]
      g__pulses[INFO_HIGHT].out_of_range++;
    }
    else if (l__delta > g__pause_min) {   // Repos détecté
      Serial.print("[");
      printWithFormat(g__infos[CURRENT].bits);
      Serial.print("] ");

      l__flg_frame = printDateAndTime(g__infos[CURRENT].bits);

      g__pause_duration = (unsigned int)l__delta;

#if USE_PRINT_STATS
      if (l__flg_frame == true) {
        printStats(INFO_FRAME_OK);
      }
      else {
        printStats(INFO_FRAME_KO);
      }
#endif

      // Mise à jour des durées des pulses et de la pause dans le cas d'une trame correcte (longueur, parités et contenu)
      if (l__flg_frame == true) {
        setPulseAndPauseDurations();
      }

      clearStats();
      Serial.print("\n");

      if (st_gestion.state == CONNECTED) {
        st_gestion.state_line = FRAME_END_DETECTED;
        if (st_gestion.duration_frame_end == -1) {
          st_gestion.duration_frame_end = 100;    // 1 Sec.
        }
      }

      if (l__flg_frame == true) {
        g__infos[CURRENT].dt.seconds++;

        /* Si 'g__infos[LAST_GOOD].available' est à false, aucune trame n'a jamais été détectée
         * => Armement du timer 'st_gestion.duration_synchro' à 30 Sec. pour se caler sur la
         *    30ème secondes de la prochaine trame (supposée correcte) et ensuite armer le
         *    timer 'st_gestion.duration_synchro' à 60 Sec. pour déterminer la dérive entre
         *    les secondes du DCF77 et comptabilisées en interne         
         *    => Dès qu'une différence apparaitra, un recalage sera effectué...
         */
        if (g__infos[LAST_GOOD].available == false) {
          /* Ne doit arriver qu'une seule fois car 'g__infos[LAST_GOOD].available'
           * n'est jamais remis à 'false'
           * TODO: Test de remise à 'false' de 'g__infos[LAST_GOOD].available' dès passage dans
           *       l'état "non connecté"
           *       => - Vérifier que 'Init synchro' est refait (armement de 30" de la seconde de synchro)
           *          - La dérive doit être moindre (à vérifier @Stats2 et au message
           *            "#Seconds for synchro #xxx (dérive: +/-yyy Sec.)"
           */
          char l__buffer[48];

          st_gestion.duration_synchro = 3000;    // 30 Sec.

          sprintf(l__buffer, "#Init. synchro\n");
          Serial.print(l__buffer);
        }

        // Sauvegarde des données de cette trame correcte
        g__infos[LAST_GOOD].timestamp = time;
        memcpy(&g__infos[LAST_GOOD].dt, &g__infos[CURRENT].dt, sizeof(ST_DATE_AND_TIME));
        g__infos[LAST_GOOD].available = true;

        /* Effacement de la disponibilité de 'g__infos[ESTIMATED]' au profit
         * de la nouvelle date détectée correctement
         */
        g__infos[ESTIMATED].available = false;

        // Nouvelle minute présentée
        g__stats2.nbr_total_minutes++;

#if USE_LEDS_FOR_TIME
        if (st_gestion.duration_for_led_hours <= 0 && st_gestion.duration_for_led_minutes <= 0) {   // Pas de réarmement si en cours
          if (g__infos[LAST_GOOD].dt.hours > 0) {
            // -1 pour le modulo 50 (2 plages [49, 48, ..., 24, 25]  et [24, 23, ..., 1, 0])
            st_gestion.duration_for_led_hours = 100 * g__infos[LAST_GOOD].dt.hours - 1;         //  ~1 Sec. x hours
          }
          if (g__infos[LAST_GOOD].dt.minutes > 0) {
            // -1 pour le modulo 50 (2 plages [49, 48, ..., 24, 25]  et [24, 23, ..., 1, 0])
            st_gestion.duration_for_led_minutes = 100 * g__infos[LAST_GOOD].dt.minutes - 1;     //  ~1 Sec. x minutes
          }

          // Calcul of Current Epoch Time
          long l__seconds = calculatedEpochTime(&g__infos[LAST_GOOD].dt);
          char l__buffer[80];

          if (g__epochTimeCurrent > 0L) {
            long l__delta = (g__epochTimeCurrent + 60L - l__seconds);
            g__epoch_time_exp_delta_min = (l__delta < g__epoch_time_exp_delta_min) ? l__delta : g__epoch_time_exp_delta_min;
            g__epoch_time_exp_delta_max = (l__delta > g__epoch_time_exp_delta_max) ? l__delta : g__epoch_time_exp_delta_max;

            //sprintf(l__buffer, "#Epoch time [%ld] Sec. (Delta: [%ld] Sec.)\n", l__seconds, l__delta);
            //Serial.print(l__buffer);

            if (l__delta != 0L) {
              if (st_gestion.error < INTERNAL_ERROR) {   // "Internal error" persistante ;-)
                st_gestion.error = INVALID_DATE_TIME_VS_ESTIMATED;
                st_gestion.nbr_for_error_pulses_init = 6;    // INVALID_DATE_TIME_VS_ESTIMATED

                sprintf(l__buffer, "#Invalid Epoch time [%ld] Sec. (Delta: [%ld] Sec.)\n", l__seconds, l__delta);
                Serial.print(l__buffer);
              }
            }
          }
          else {
            sprintf(l__buffer, "#Epoch time [%ld] Sec.\n", l__seconds);
            Serial.print(l__buffer);

            g__epochTimeInit = l__seconds;          // 1st Epoch Time calculated
          }

          g__epochTimeCurrent = l__seconds;
        }
#endif
      }

      // Préparation pour une prochaine acquisition
      g__infos[CURRENT].timestamp = (unsigned long)-1;
      memset(&g__infos[CURRENT].dt, '\0', sizeof(ST_DATE_AND_TIME));
      g__infos[CURRENT].dt.seconds = -1;
      g__infos[CURRENT].bits = "";
      g__infos[CURRENT].available = false;
    }
    else {
      /* Pulse -\__/-- détectée de durée <= 'g__pause_min'
       *  => Non significatif compte tenu des variations des pulses _/--\__ (bits '0' et '1')
       */

    }
  }
  else if (pulse == 0) {
    // Pulse _/--\__ détectée de durée <= 'g__pulse_0_min'
    g__pulses[INFO_HIGHT].out_of_range++;
  }
  // Fin: Traitement sur les durées uniquement si transition détectée

  if (g__infos[CURRENT].bits.length() > FRAME_LENGTH) {
      Serial.print("[");
      printWithFormat(g__infos[CURRENT].bits);
      Serial.print("] Error: Frame too long");

      g__infos[CURRENT].bits = "";

      if (st_gestion.error < INTERNAL_ERROR) {   // "Internal error" persistante ;-)
        st_gestion.error = FRAME_TOO_LONG;
        st_gestion.nbr_for_error_pulses_init = 4;   // FRAME_TOO_LONG
      }

#if USE_PRINT_STATS
      printStats(INFO_FRAME_KO);
#endif
      clearStats();
      Serial.print("\n");
  }

more:
  // Gestion des états, des erreurs et des timers
  statesGestion(time);
  // Fin: Gestion des états, des erreurs et des timers

#if USE_PRINT_STATS2
  if (g__stats2.flg_report == true) {
    printStats2();
    g__stats2.flg_report = false;   // Effacement
  }
#endif

  // Recopie physique des états Leds
  digitalWrite(LED_RED,    (g__state_leds & STATE_LED_RED) ? HIGH : LOW);
  digitalWrite(LED_YELLOW, (g__state_leds & STATE_LED_YELLOW) ? HIGH : LOW);
  digitalWrite(LED_GREEN,  (g__state_leds & STATE_LED_GREEN) ? HIGH : LOW);

#if USE_GET_SERIAL_IN
  if (st_gestion.state == NOT_CONNECTED) {
    if (Serial.available() > 0) {
      char l__buffer[96];

      // Read the incoming byte
      int incomingByte = Serial.read();

      if (incomingByte == '\n') {
        g__incoming_buff[g__count] = '\0';
        g__count = 0;

        sprintf(l__buffer, ">>> [%s] (length: %d)\n", g__incoming_buff, strlen(g__incoming_buff));
        Serial.print(l__buffer);

        size_t n = 0;
        boolean l__flg_err = false;
        String l__frame = "";
        for (n = 0; n < strlen(g__incoming_buff); n++) {
          if (g__incoming_buff[n] != '0' && g__incoming_buff[n] != '1' && g__incoming_buff[n] != '-') {
            sprintf(l__buffer, ">>> #%d: Invalid character (%c)\n", n, g__incoming_buff[n]);
            Serial.print(l__buffer);
            l__flg_err = true;
            break;
          }

          if (g__incoming_buff[n] != '-') {
            l__frame += g__incoming_buff[n];
          }
        }
        if (l__flg_err == false) {
          Serial.print(">>> ");
          l__flg_err = printDateAndTime(l__frame);
          Serial.print("\n");

          if (l__flg_err == true) {
            // Calcul of next minute
            memcpy(&g__infos[ESTIMATED].dt, &g__infos[CURRENT].dt, sizeof(ST_DATE_AND_TIME));
            g__infos[LAST_GOOD].available = false;
            g__infos[ESTIMATED].available = true;

            calculOfDateAndTime();

            Serial.print("<<< [");
            printDateAndTime();
            Serial.print("] (Date and time calculated)\n");

            long l__seconds = calculatedEpochTime(&g__infos[CURRENT].dt);

            sprintf(l__buffer, "<<< Epoch timestamp [%ld] Sec. since 01/01/1970\n", l__seconds);
            Serial.print(l__buffer);
          }
        }
      }
      else if (g__count < sizeof(g__incoming_buff)){
        g__incoming_buff[g__count++] = (char)incomingByte;
      }
      else {
        sprintf(l__buffer, ">>> Buffer overflow (%d >= %d)\n", g__count, sizeof(g__incoming_buff));
        Serial.print(l__buffer);
        g__count = 0;
      }
    }
  }
#endif
}
